package com.bocloud.fms.booter.controller;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.intf.FlowDefinitionService;
import com.bocloud.fms.model.BasicDomainModel;
import com.bocloud.fms.model.DefinitionParams;
import com.bocloud.fms.model.DefinitionPatchParam;
import com.megatron.common.model.*;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.GridHelper;
import com.megatron.common.utils.MapTools;
import com.megatron.common.utils.ResultTools;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.web.bind.annotation.*;

import java.util.List;

/**
 * @author hufy
 */
@RestController
@Tag(name = "流程定义")
@RequestMapping({"/v1/definitions"})
@Slf4j
public class FlowDefinitionController {

    @Autowired
    protected FlowDefinitionService definitionService;

    /**
     * 获取流程定义列表
     *
     * @param pager   分页过滤参数
     * @param context 请求上下文
     * @param full    是否获取全量数据
     * @return 分页数据
     */
    @GetMapping
    @Operation(summary = "流程定义列表")
    public GeneralResult list(Pager pager, @Value(Common.REQ_CONTEXT) RequestContext context, @RequestParam(value = "full", required = false) Boolean full) {
        if (null != pager.getParams()) {
            pager.getParams().add(new Param(MapTools.simpleMap("tenantId", context.getTenant()), Sign.EQ));
        }
        if (null != full && full) {
            pager.setRows(Integer.MAX_VALUE);
        }
        int count = definitionService.count(context, pager, true);
        List<FlowDefinition> latestDefinitions = definitionService.getDefinitions(context, pager, true);
        GridBean gridBean = GridHelper.getBean(pager.getPage(), pager.getRows(), count, latestDefinitions);
        return new GeneralResult(true, gridBean, "成功");
    }

    /**
     * 新增流程定义
     *
     * @param context 请求上下文
     * @param params  流程定义实体类
     * @return 统一返回参数 - 是否创建成功
     */
    @PostMapping
    @Operation(summary = "新增流程定义")
    public GeneralResult create(@Value(Common.REQ_CONTEXT) RequestContext context,
                                @RequestBody DefinitionParams params) {
        log.info("修改流程定义 {}", JSONObject.toJSONString(params));
        Long definitionId = definitionService.deploy(params, context);
        if (definitionId == null) {
            return new GeneralResult(false, "创建失败");
        }
        BasicDomainModel<FlowDefinition, DefinitionModel> definitionById = definitionService
                .getDefinition(definitionId);
        if (definitionById != null) {
            return new GeneralResult(true, "创建成功");
        } else {
            return new GeneralResult(false, "创建失败");
        }
    }

    /**
     * 修改流程定义
     *
     * @param id     被修改的流程id
     * @param params 流程定义实体类
     * @return 统一返回参数 - 是否修改成功
     */
    @PutMapping(value = "/{id}")
    @Operation(summary = "修改流程定义")
    public GeneralResult modify(@Value(Common.REQ_CONTEXT) RequestContext context, @PathVariable(Common.ID) Long id,
                                @RequestBody DefinitionParams params) {
        return definitionService.redeploy(id, params, context);
    }

    /**
     * 激活、终止流程定义
     *
     * @param context 请求上下文
     * @param id      激活或者终止流程id
     * @param param   流程定义实体类
     * @return 统一返回参数 - 是否激活成功
     */
    @PatchMapping(value = "/{id}")
    @Operation(summary = "激活、终止流程定义")
    public GeneralResult patch(@Value(Common.REQ_CONTEXT) RequestContext context, @PathVariable(Common.ID) Long id,
                               @RequestBody DefinitionPatchParam param) {
        if ("terminate".equals(param.getAction())) {
            boolean success = definitionService.terminate(id, context, param.getRemark());
            return new GeneralResult(success, success ? "终止成功" : "终止失败");
        } else if ("active".equals(param.getAction())) {
            boolean success = definitionService.active(id, context, param.getRemark());
            return new GeneralResult(success, success ? "激活成功" : "激活失败");
        } else {
            return ResultTools.formatErrResult();
        }
    }

    /**
     * 删除流程定义
     *
     * @param id 被删除的流程id
     * @return 统一返回参数 - 是否删除成功
     */
    @DeleteMapping(value = "/{id}")
    @Operation(summary = "删除流程定义")
    public GeneralResult remove(@Value(Common.REQ_CONTEXT) RequestContext context, @PathVariable(Common.ID) Long id) {
        return definitionService.remove(id, context);
    }

    /**
     * 流程定义详情
     *
     * @param id 获取流程id
     * @return 流程定义实体类
     */
    @GetMapping(value = "/{id}")
    @Operation(summary = "流程定义详情")
    public GeneralResult detail(@PathVariable(Common.ID) Long id) {
        BasicDomainModel<FlowDefinition, DefinitionModel> definition = definitionService.getDefinition(id);
        if (definition == null) {
            return new GeneralResult(false, "没有找到该流程定义");
        }
        FlowDefinition entity = definition.getEntity();
        if (entity != null) {
            return new GeneralResult(true, entity, "Success");
        } else {
            return new GeneralResult(false, "Failed");
        }
    }

    /**
     * 查看更新历史
     *
     * @param id 查看的更新历史流程id
     * @return 更新历史
     */
    @GetMapping(value = "/{id}/history")
    @Operation(summary = "查看更新历史")
    public GeneralResult listHistory(@Value(Common.REQ_CONTEXT) RequestContext context,
                                     @PathVariable(Common.ID) Long id) {
        List<FlowDefinition> definitionHistory = definitionService.getDefinitionHistory(context, id);
        return new GeneralResult(true, definitionHistory, "Success");
    }

}
