package com.bocloud.fms.booter.controller;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.fms.engine.AssignmentHandler;
import com.bocloud.fms.engine.FlowEngine;
import com.bocloud.fms.engine.model.AuditTaskModel;
import com.bocloud.fms.engine.model.BaseNodeModel;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.engine.model.FlowNodeModel;
import com.bocloud.fms.entity.Attachment;
import com.bocloud.fms.entity.FlowInstance;
import com.bocloud.fms.entity.FlowNode;
import com.bocloud.fms.enums.FlowConstants;
import com.bocloud.fms.intf.AttachmentService;
import com.bocloud.fms.intf.FlowQueryService;
import com.bocloud.fms.model.BasicDomainModel;
import com.bocloud.fms.model.FlowInstancePatchParam;
import com.bocloud.fms.service.Execution;
import com.google.common.collect.Lists;
import com.megatron.common.model.*;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.web.bind.annotation.*;

import java.util.*;

/**
 * @author hufy
 */
@Slf4j
@RequiredArgsConstructor
@RestController
@Tag(name = "流程实例")
@RequestMapping({"/v1/instances"})
public class FlowInstanceController {

    private final FlowEngine engine;
    private final AssignmentHandler assignmentHandler;
    private final AttachmentService attachmentService;
    private final FlowQueryService flowQueryService;

    /**
     * 展示申请的或审核过的表单列表
     *
     * @param pager   分页参数
     * @param type    查询的类型，已审批，待审批，审批结束和所有
     * @param context 请求上下文
     * @return 流程示例列表
     */
    @GetMapping
    @Operation(summary = "流程实例列表")
    public GeneralResult<?> list(Pager pager, String type, @Value(Common.REQ_CONTEXT) RequestContext context) {
        pager.setParams(Optional.ofNullable(pager.getParams()).orElse(new ArrayList<>()));
        if (StringUtils.isBlank(type)) {
            return new GeneralResult<>(false, "重要参数为空");
        }
        pager.getParams().forEach(param -> {
            if (null != param.getParam().get("name") && StringUtils.isNotBlank((String) param.getParam().get("name"))) {
                param.getParam().put("o.name", param.getParam().get("name"));
                param.getParam().remove("name");
            }
        });
        return engine.query().getGridBeanOfFlowInstances(pager, context, type);
    }

    /**
     * 启动一个流程
     *
     * @param id             要启动的流程id
     * @param requestContext 请求上下文
     * @param params         流程启动参数
     * @param context        请求上下文
     * @param project        项目
     * @return 统一返回参数 - 流程是否启动成功
     */
    @PostMapping
    @Operation(summary = "启动流程实例")
    @Deprecated
    public GeneralResult<Long> start(@RequestParam("defId") Long id, @Value(Common.REQ_CONTEXT) RequestContext requestContext,
                                     @RequestParam(value = Common.PARAMS) String params,
                                     @RequestParam(value = Common.CONTEXT, required = false) String context,
                                     @RequestParam(value = "project", required = false) String project) {
        Execution execution;
        Map<String, Object> map = JSONObject.parseObject(params);
        // 如果为远程调用，则context不为null 远程调用返回和http返回参数不通，故分别返回
        if (null != context) {
            requestContext = JSONObject.parseObject(context, RequestContext.class);
        }
        execution = engine.startInstanceById(id, requestContext, Long.parseLong(project), map);
        if (null != execution) {
            return new GeneralResult<>(true, execution.getFlowInstance().getId(), "启动流程成功");
        }
        return new GeneralResult<>(false, "启动流程失败");
    }

    /**
     * 撤销流程
     *
     * @param id         要撤销的流程id
     * @param patchParam 撤销流程参数
     * @param context    请求上下文
     * @return 统一返回参数 - 是否撤销成功
     */
    @PatchMapping(value = "/{id}")
    @Operation(summary = "撤销流程实例")
    public GeneralResult<Void> patch(@PathVariable(Common.ID) Long id, @RequestBody FlowInstancePatchParam patchParam,
                                     @Value(Common.REQ_CONTEXT) RequestContext context) {
        if ("terminate".equals(patchParam.getAction())) {
            boolean success = engine.flowInstance().cancel(id, context.getTarget(), null, true);
            return new GeneralResult<>(success, success ? "流程撤销成功" : "流程撤销失败");
        } else {
            return new GeneralResult<>(false, "参数格式错误", "400", "修改参数格式");
        }
    }

    /**
     * 流程实例详情
     *
     * @param id 流程源id
     * @return 流程实体类
     */
    @GetMapping(value = "/{id}")
    @Operation(summary = "流程实例详情")
    public GeneralResult<FlowInstance> detail(@PathVariable(Common.ID) Long id) {
        FlowInstance flowInstance = engine.query().getFlowInstance(id);
        if (flowInstance == null) {
            return new GeneralResult<>(false, "未查询到数据");
        } else {
            return new GeneralResult<>(true, flowInstance, "Success!");
        }
    }

    /**
     * 流程实例审核节点列表
     *
     * @param id      流程实例id
     * @param context 请求上下文
     * @param type    类型，可选值：audit、record
     * @return 流程实例审核节点列表
     */
    @GetMapping(value = "/{id}/nodes")
    @Operation(summary = "流程实例审核节点列表")
    public GeneralResult<List<FlowNode>> nodes(@PathVariable(Common.ID) Long id, @Value(Common.REQ_CONTEXT) RequestContext context,
                                               @RequestParam(value = "type") String type) {
        if ("audit".equals(type)) {
            List<Param> params = new ArrayList<>();
            params.add(new Param(MapTools.simpleMap("t.instanceId", id), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("instanceState", FlowConstants.STATE_ACTIVE), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("nodeState", FlowConstants.STATE_ACTIVE), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("actorId", context.getTarget()), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("actorCatalog", RequestContext.Catalog.Tenant == context.getCatalog() ? RequestContext.Catalog.Tenant : RequestContext.Catalog.User), Sign.EQ));
            // 获取流程定义id
            FlowInstance flowInstance = engine.query().getFlowInstance(id);
            DefinitionModel definitionModel = engine.definition().getDefinition(flowInstance.getDefId()).getModel();
            Set<RequestContext> userIds = new HashSet<>();
            Map<String, Set<RequestContext>> curNameAndNextActors = new HashMap<>();

            List<FlowNode> list = Lists.newArrayList();
            List<BasicDomainModel<FlowNode, FlowNodeModel>> basicDomainModels = engine.query().getActiveNodes(params);
            for (BasicDomainModel<FlowNode, FlowNodeModel> basicModel : basicDomainModels) {
                FlowNode entity = basicModel.getEntity();
                BaseNodeModel currentNode = definitionModel.getNode(entity.getNodeId());
                if (null == currentNode) {
                    continue;
                }
                List<AuditTaskModel> nextModels = currentNode.getNextModels(AuditTaskModel.class);
                Set<RequestContext> nextActors = new HashSet<>();
                if (!ListTool.isEmpty(nextModels)) {
                    for (AuditTaskModel nextModel : nextModels) {
                        nextActors = assignmentHandler.getAssignees(nextModel, flowInstance.getProjectId(), context);
                    }
                    userIds.addAll(nextActors);
                }
                curNameAndNextActors.put(currentNode.getId(), nextActors);
                list.add(entity);
            }

            // 查询用户名称
            if (userIds.size() > 0 && !ListTool.isEmpty(list)) {
                for (FlowNode flowNode : list) {
                    Set<RequestContext> nextActors = curNameAndNextActors.get(flowNode.getNodeId());
                    StringBuilder names = new StringBuilder();
                    String actorNames = null;
                    for (RequestContext actorContext : nextActors) {
                        String name = flowQueryService.getDisplayName(actorContext);
                        names.append(name).append(",");
                    }
                    if (names.length() > 0) {
                        actorNames = names.substring(0, names.lastIndexOf(","));
                    }
                    flowNode.setNextActorNames(actorNames);
                }
            }
            return new GeneralResult<>(true, list, "Success");
        } else if ("record".equals(type)) {
            FlowInstance flowInstance = engine.query().getFlowInstance(id);
            List<Param> params = new ArrayList<>();
            params.add(new Param(MapTools.simpleMap("t.instanceId", id), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("t.gmtFinish", "1970-1-1"), Sign.GT));

            List<FlowNode> nodes;
            // 判断用户是否创建此流程
            Map<String, String> sorter = new LinkedHashMap<>();
            sorter.put("gmtCreate", Common.ONE);
            sorter.put("gmtFinish", Common.ONE);
            sorter.put(Common.ID, Common.ONE);
            if (flowInstance.getCreatorId().equals(context.getTarget())) {
                nodes = engine.query().getNodes(context, new Pager(1, Integer.MAX_VALUE, params, sorter));
            } else {
                // 用户是否参与此流程
                params.add(new Param(MapTools.simpleMap("actorId", context.getTarget()), Sign.EQ));
                // 查询有当前参与者的节点
                nodes = engine.query().getNodes(context, new Pager(1, Integer.MAX_VALUE, params, sorter));
                // 如果不为空，说明有参与，查询所有的节点
                if (nodes != null) {
                    params = new ArrayList<>();
                    params.add(new Param(MapTools.simpleMap("t.instanceId", id), Sign.EQ));
                    params.add(new Param(MapTools.simpleMap("t.gmtFinish", "1970-1-1"), Sign.GT));
                    nodes = engine.query().getNodes(context, new Pager(1, Integer.MAX_VALUE, params, sorter));
                }
            }
            if (!ListTool.isEmpty(nodes)) {
                for (FlowNode node : nodes) {
                    if (FlowConstants.STATE_ACTIVE.equals(node.getNodeState())) {
                        node.setVariable(null);
                        if (FlowConstants.STATE_CANCEL.equals(flowInstance.getInstanceState())) {
                            // 给前端用的，前端希望知道活动中的node所在流程是否撤销
                            node.setNodeState(FlowConstants.STATE_CANCEL);
                        }
                    } else if (FlowConstants.STATE_FINISH.equals(node.getNodeState())) {
                        // 给前端用的，前端希望知道node是否审核通过,好做展示
                        node.setNodeState((String) node.getVariableMap().get(FlowConstants.ARGS_AUDIT_FLAG));
                    }
                    if (StringUtils.isNotBlank(node.getGroupId())) {
                        List<Attachment> attachments = attachmentService.list(node.getGroupId());
                        attachments.forEach(o -> {
                            if (!o.getPath().startsWith("/")) {
                                o.setPath("/" + o.getPath());
                            }
                        });
                        node.setAttachments(attachments);
                    }
                }
                return new GeneralResult<>(true, nodes, "Success");
            }
            return new GeneralResult<>(false, "未查询到数据");
        } else {
            return new GeneralResult<>(false, "参数格式错误", "400", "修改参数格式");
        }
    }

    /**
     * 获取流程工单数量
     *
     * @param context 请求上下文
     * @return 流程工单的数量
     */
    @GetMapping("/count")
    @Operation(summary = "获取流程工单数量")
    public GeneralResult<Long> countTodo(String context) {
        log.debug("接收获取流程工单数量请求: {} ", JSONObject.toJSONString(context));
        try {
            long countFlowInstancesTodo = flowQueryService.countTodo(JSONObject.parseObject(context, RequestContext.class));
            return new GeneralResult<>(true, countFlowInstancesTodo, "获取流程工单数量成功");
        } catch (Exception e) {
            return new GeneralResult<>(false, "获取流程工单数量失败!");
        }
    }

}
