package com.bocloud.fms.service;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudExchange;
import com.bocloud.fms.enums.FlowConstants;
import com.bocloud.fms.exception.FlowException;
import com.bocloud.fms.model.BasicDomainModel;
import com.bocloud.fms.utils.AbstractAssertUtil;
import com.bocloud.fms.engine.AssignmentHandler;
import com.bocloud.fms.engine.Expression;
import com.bocloud.fms.engine.FlowEngine;
import com.bocloud.fms.engine.cache.CacheManager;
import com.bocloud.fms.engine.cache.CacheManagerAware;
import com.bocloud.fms.engine.cache.memory.MemoryCacheManager;
import com.bocloud.fms.engine.cfg.FlowConfiguration;
import com.bocloud.fms.engine.model.*;
import com.bocloud.fms.engine.spring.SpelExpression;
import com.bocloud.fms.entity.Attachment;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.entity.FlowInstance;
import com.bocloud.fms.entity.FlowNode;
import com.bocloud.fms.intf.*;
import com.bocloud.fms.repository.FlowNodeRepository;
import com.bocloud.fms.service.internal.FmsSmsUserService;
import com.google.common.collect.Maps;
import com.megatron.common.model.*;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.AmqpException;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import java.util.*;

/**
 * 基本的流程引擎实现类
 *
 * @author hufy
 */
@Slf4j
public class FlowEngineImpl implements FlowEngine {
    /**
     * Flow配置对象
     */
    protected FlowConfiguration flowConfiguration;
    /**
     * 流程定义业务类
     */
    @Autowired
    protected FlowDefinitionService definitionService;
    /**
     * 流程实例业务类
     */
    @Autowired
    protected FlowInstanceService flowInstanceService;
    /**
     * 任务业务类
     */
    @Autowired
    protected FlowNodeService flowNodeService;
    /**
     * 查询业务类
     */
    @Autowired
    protected FlowQueryService queryService;
    @Autowired
    protected FmsSmsUserService fmsSmsUserService;
    @Autowired
    protected AssignmentHandler assignmentHandler;
    @Autowired
    protected AmqpTemplate amqpTemplate;
    @Autowired
    protected AttachmentService attachmentService;
    @Autowired
    protected FlowNodeRepository flowNodeRepository;

    private Expression expression = new SpelExpression();

    /**
     * 根据serviceContext上下文，查找definitionService、flowInstanceService、taskService服务
     */
    @Override
    public FlowEngine configure(FlowConfiguration config) {
        this.flowConfiguration = config;
        definitionService = AbstractServiceContext.find(FlowDefinitionService.class);
        queryService = AbstractServiceContext.find(FlowQueryService.class);
        flowInstanceService = AbstractServiceContext.find(FlowInstanceService.class);
        flowNodeService = AbstractServiceContext.find(FlowNodeService.class);

        CacheManager cacheManager = AbstractServiceContext.find(CacheManager.class);
        if (cacheManager == null) {
            // 默认使用内存缓存管理器
            cacheManager = new MemoryCacheManager<>();
        }
        List<CacheManagerAware> cacheServices = AbstractServiceContext.findList(CacheManagerAware.class);
        for (CacheManagerAware cacheService : cacheServices) {
            cacheService.setCacheManager(cacheManager);
        }
        return this;
    }

    /**
     * 获取流程定义服务
     */
    @Override
    public FlowDefinitionService definition() {
        AbstractAssertUtil.notNull(definitionService);
        return definitionService;
    }

    /**
     * 获取查询服务
     */
    @Override
    public FlowQueryService query() {
        AbstractAssertUtil.notNull(queryService);
        return queryService;
    }

    /**
     * 获取实例服务
     *
     * @since 1.2.2
     */
    @Override
    public FlowInstanceService flowInstance() {
        AbstractAssertUtil.notNull(flowInstanceService);
        return flowInstanceService;
    }

    /**
     * 获取任务服务
     *
     * @since 1.2.2
     */
    @Override
    public FlowNodeService node() {
        AbstractAssertUtil.notNull(flowNodeService);
        return flowNodeService;
    }

    @Override
    public AssignmentHandler assignmentHandler() {
        return assignmentHandler;
    }

    /**
     * 根据流程定义ID，操作人ID，参数列表启动流程实例
     */
    @Override
    @Transactional
    public Execution startInstanceById(Long id, RequestContext context, Long project, Map<String, Object> args) {
        args = Optional.ofNullable(args).orElse(new HashMap<>());
        BasicDomainModel<FlowDefinition, DefinitionModel> definition = definition().getLatestDefinition(id);
        Assert.notNull(definition, "未找到该流程");
        Assert.isTrue(FlowConstants.STATE_ACTIVE.equals(definition.getEntity().getState()), "该流程定义已删除");
        Assert.isTrue(!definition.getEntity().getDeleted(), "该流程或已终止");
//        checkFlowUserStatus(definition, context, project);
        definition().check(definition.getEntity(), id);
        return startDefinition(definition, context, project, args);
    }

    private void checkFlowUserStatus(BasicDomainModel<FlowDefinition, DefinitionModel> definition, RequestContext context, Long project) {
        // 判断流程中的用户是否还存在
        List<BaseNodeModel> nodes = definition.getModel().getNodes();
        Assert.isTrue(!ListTool.isEmpty(nodes), "流程定义错误，请联系管理员");
        for (BaseNodeModel node : nodes) {
            if (node instanceof AuditTaskModel) {
                Set<RequestContext> actors = assignmentHandler.getAssignees((AuditTaskModel) node, project, context);
                //没有人时自动审批通过
//                Assert.isTrue(actors != null && actors.size() > 0, "流程审批人定义错误，未找到下一环节审批人，请联系管理员");
            }

        }
    }

    private Execution startDefinition(BasicDomainModel<FlowDefinition, DefinitionModel> definitionDomain,
                                      RequestContext context, Long project, Map<String, Object> args) {
        Execution execution = execute(definitionDomain, context, project, args);
        Assert.notNull(definitionDomain.getModel(), "No model");
        if (definitionDomain.getModel() != null) {
            StartModel start = definitionDomain.getModel().getStart();
            AbstractAssertUtil.notNull(start, "流程定义[name=" + definitionDomain.getEntity().getName() + ", version="
                    + definitionDomain.getEntity().getVersion() + "]没有开始节点");
            start.execute(execution);
        }
        return execution;
    }

    /**
     * 创建流程实例，并返回执行对象
     *
     * @param definition 流程定义
     * @param context    操作人
     * @param project    项目ID
     * @param args       参数列表
     */
    private Execution execute(BasicDomainModel<FlowDefinition, DefinitionModel> definition, RequestContext context,
                              Long project, Map<String, Object> args) {
        FlowInstance flowInstance = flowInstance().createFlowInstance(definition, context, project, args, null, null);
        if (log.isDebugEnabled()) {
            log.debug("创建流程实例对象:" + flowInstance);
        }
        Execution current = new Execution(this, definition, flowInstance, args);
        current.setContext(context);
        return current;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public GeneralResult batchExecuteNode(List<Long> instanceIds, RequestContext context, Map<String, Object> args,
                                          List<Attachment> attachments) {
        List<Long> taskIds = new ArrayList<>(instanceIds.size());
        for (Long instanceId : instanceIds) {
            FlowInstance flowInstance = this.flowInstance().getFlowInstance(instanceId);
            List<Param> params = new ArrayList<>();
            params.add(new Param(MapTools.simpleMap("t.instanceId", instanceId), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("instanceState", FlowConstants.STATE_ACTIVE), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("nodeState", FlowConstants.STATE_ACTIVE), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("actorId", context.getTarget()), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("actorCatalog", context.getCatalog() == RequestContext.Catalog.Tenant ? RequestContext.Catalog.Tenant : RequestContext.Catalog.User), Sign.EQ));
            List<BasicDomainModel<FlowNode, FlowNodeModel>> basicDomainModels = this.query().getActiveNodes(params);
            if (basicDomainModels.size() > 1) {
                return new GeneralResult(false, "申请：" + flowInstance.getInstanceNo() + "存在多个审核节点无法进行批量审批操作");
            }
            if (basicDomainModels.size() == 0) {
                log.warn("batch audit warning, this instance without audit node, ignore! ");
            }
            taskIds.add(basicDomainModels.get(0).getEntity().getId());
        }
        for (Long taskId : taskIds) {
            this.executeNode(taskId, context, args, attachments);
        }
        return new GeneralResult(true, "审核成功");
    }

    /**
     * 根据任务主键ID，操作人ID，参数列表执行任务
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public GeneralResult executeNode(Long taskId, RequestContext context, Map<String, Object> variable,
                                     List<Attachment> attachments) {
        Map<String, Object> args = new HashMap<>();
        args.putAll(Optional.ofNullable(variable).orElse(Maps.newHashMap()));
        boolean isDraft = args.get(FlowConstants.ARGS_IS_DRAFT) != null
                && Boolean.valueOf(String.valueOf(args.get(FlowConstants.ARGS_IS_DRAFT)));
        args.put(FlowConstants.ARGS_IS_DRAFT, args.get(FlowConstants.ARGS_IS_DRAFT));
        FlowNode node = node().getFlowNode(taskId);
        FlowInstance flowInstance = flowInstance().getFlowInstance(node.getInstanceId());
        // 保存附件
        saveAttachments(attachments, node, context);
        if (isDraft) {// 暂存
            node.setVariable(JSONObject.toJSONString(args));
            flowInstance.getVariableMap().putAll(args);
            flowInstance.setVariable(JSONObject.toJSONString(flowInstance.getVariableMap()));
            flowInstance.setGmtUpdate(new Date());
            flowInstance.setMenderId(context.getTarget());
            flowInstance.setMenderCatalog(context.getCatalog());
            Long projectId = Long
                    .valueOf(String.valueOf(flowInstance.getVariableMap().get(FlowConstants.ARGS_PROJECT_ID)));
            flowInstance.setProjectId(projectId);
            flowInstance().updateFlowInstance(flowInstance);
            Map<String, Object> map = Maps.newHashMap();
            map.put("instanceId", flowInstance.getId());
            map.put("taskId", taskId);
            return new GeneralResult(true, map, "保存成功");
        }
        // 检查下一下环节审批人是否存在
        //没有审批人，自动通过，这里不再校验
//        checkNextAuditUser(context, flowInstance, node, args);
        // 完成任务，并且构造执行对象
        Execution execution = execute(taskId, context, args);
        // sonar : 下方注释代码永远不会执行
//        if (execution == null) {
//            return new GeneralResult(true, Collections.emptyList(), "审核成功");
//        }
        DefinitionModel model = execution.getDefinition().getModel();
        if (model != null) {
            BaseNodeModel nodeModel = model.getNode(execution.getTask().getEntity().getNodeId());
            // 将执行对象交给该任务对应的节点模型执行
            nodeModel.execute(execution);
        }
        //发送websocket消息用于前端更新内容
        siteMessage(context);

        flowInstance = flowInstance().getFlowInstance(node.getInstanceId());
        Map<String, Object> map = Maps.newHashMap();
        map.put("instanceId", flowInstance.getId());
        return new GeneralResult(true, map, "审核成功");
    }


    /**
     * 保存附件信息
     *
     * @param attachments
     * @param node
     * @param context
     */
    private void saveAttachments(List<Attachment> attachments, FlowNode node, RequestContext context) {
        if (attachments == null || attachments.size() == 0) {
            return;
        }
        String groupId = StringUtils.isBlank(node.getGroupId()) ? UUID.randomUUID().toString().replace("-", "")
                : node.getGroupId();
        for (Attachment attachment : attachments) {
            attachment.setGroupId(groupId);
            attachment.setCreatorId(context.getTarget());
            attachment.setGmtCreate(new Date());
        }
        if (!groupId.equals(node.getGroupId())) {
            node.setGroupId(groupId);
            node().updateTask(node);
        }
        attachmentService.saveAttachment(attachments);
    }

    /**
     * 预检查下一步执行人是否存在，不存在则抛出异常
     *
     * @param flowInstance 流程实例
     * @param node         审核节点
     * @param args         流程变量
     * @throws FlowException 异常
     */
    private void checkNextAuditUser(RequestContext context, FlowInstance flowInstance, FlowNode node,
                                    Map<String, Object> args) throws FlowException {
        Map<String, Object> variableMap = flowInstance.getVariableMap();
        variableMap.putAll(args);

        BasicDomainModel<FlowDefinition, DefinitionModel> definition = definition()
                .getDefinition(flowInstance.getDefId());
        BaseNodeModel nodeModel = definition.getModel().getNode(node.getNodeId());
        for (TransitionModel tm : nodeModel.getOutputs()) {
            String auditFlag = tm.getAuditFlag();
            if (StringUtils.isNotEmpty(auditFlag)
                    && auditFlag.equalsIgnoreCase((String) args.get(FlowConstants.ARGS_AUDIT_FLAG))) {
                boolean result = true;
                if (StringUtils.isNotEmpty(tm.getExpression())) {
                    result = expression.eval(Boolean.class, tm.getExpression(), args);
                }
                if (result) {
                    if (tm.getTarget() instanceof AuditTaskModel) {
                        AuditTaskModel auditTaskModel = (AuditTaskModel) tm.getTarget();
                        Set<RequestContext> actors = assignmentHandler.getAssignees(auditTaskModel, flowInstance.getProjectId(), context);
                        Assert.isTrue(actors != null && actors.size() > 0, "流程审批人定义错误，未找到下一环节审批人，请联系管理员");
                    }
                    return;
                }
            }
        }
    }

    /**
     * 根据任务主键ID，操作人ID，参数列表完成任务，并且构造执行对象
     *
     * @param taskId  任务id
     * @param context 操作人
     * @param args    参数列表
     * @return Execution
     */
    private Execution execute(Long taskId, RequestContext context, Map<String, Object> args) {
        args = Optional.ofNullable(args).orElse(new HashMap<>());
        FlowNode node = this.flowNodeRepository.getNode(taskId);
        // 更新流程变量
        FlowInstance flowInstance = query().getFlowInstance(node.getInstanceId());
        Assert.notNull(flowInstance, "指定的流程实例[id=" + node.getInstanceId() + "]已完成或不存在");
        Map<String, Object> variable = new HashMap<>();
        variable.putAll(args);
        variable.remove("id");
        variable.remove("auditFlag");
        variable.remove("remark");
        flowInstance.getVariableMap().putAll(variable);
        flowInstance.setVariable(JSONObject.toJSONString(flowInstance.getVariableMap()));
        flowInstance.setMenderId(context.getTarget());
        flowInstance.setMenderCatalog(context.getCatalog());
        flowInstance.setGmtUpdate(new Date());
        flowInstance().updateFlowInstance(flowInstance);

        args.putAll(variable);
        // 完成任务
        BasicDomainModel<FlowNode, FlowNodeModel> task = node().complete(taskId, context, args);
        Map<String, Object> flowInstanceMaps = flowInstance.getVariableMap();
        if (flowInstanceMaps != null) {
            for (Map.Entry<String, Object> entry : flowInstanceMaps.entrySet()) {
                if (args.containsKey(entry.getKey())) {
                    continue;
                }
                args.put(entry.getKey(), entry.getValue());
            }
        }
        // 查询当前版本的定义
        BasicDomainModel<FlowDefinition, DefinitionModel> definition = definition()
                .getDefinition(flowInstance.getDefId());
        Execution execution = new Execution(this, definition, flowInstance, args);
        execution.setContext(context);
        execution.setTask(task);
        return execution;
    }

    /**
     * 发送websocket消息用于前端更新内容
     *
     * @param context
     */
    private void siteMessage(RequestContext context) {
        try {
            OperateResult msgResult = new OperateResult(true, "状态变更成功", "message.change", null,
                    OperateResult.OperateCategory.SiteMessage, context);
            this.amqpTemplate.convertAndSend(BoCloudExchange.Default.TOPIC.getExchange(), "bocloud.message.queue.#",
                    msgResult);
        } catch (AmqpException e) {
            log.error(e.getMessage());
        }
    }

}
